# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pennylane as qml
import numpy as np
import jax
import jax.numpy as jnp
import optax
from sklearn.base import BaseEstimator, ClassifierMixin
from qcml.utils.model import train
from qcml.utils.model import chunk_vmapped_fn

jax.config.update("jax_enable_x64", True)


class TreeTensorClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        learning_rate=0.01,
        batch_size=32,
        max_steps=10000,
        convergence_interval=200,
        random_state=42,
        max_vmap=None,
        jit=True,
        scaling=1.0,
        dev_type="default.qubit.jax",
        qnode_kwargs={"interface": "jax-jit"},
    ):
        r"""
        Tree tensor network classifier from https://arxiv.org/abs/2011.06258v2 (see figure 1)

        This is a variational model where data is amplitude embedded and a trainiable circuit with a tree like
        structure is used for prediction. Due to the tree structure, the number of qubits must be a power of 2.

        In the plots, the data encoding state for a given input x is approximated  using a variational circuit.
        In practice, this means one has to train an encoding circuit for every training and test
        input. Since this is very expensive, we just use the exact amplitude encoded state. If the input data
        dimension is smaller than the state vector dimension, we pad with a value :math:`1/2^{n_qubits}`.

        The classification is performed via a Z measurement on the first qubit plus a trainable bias. Training
        is via the square loss.

        Args:
            learning_rate (float): Initial learning rate for gradient descent.
            batch_size (int): Size of batches used for computing parameter updates.
            convergence_interval (int): The number of loss values to consider to decide convergence.
            max_steps (int): Maximum number of training steps. A warning will be raised if training did not converge.
            random_state (int): Seed used for pseudorandom number generation
            max_vmap (int or None): The maximum size of a chunk to vectorise over. Lower values use less memory.
                must divide batch_size.
            jit (bool): Whether to use just in time compilation.
            scaling (float): Factor by which to scale the input data.
            dev_type (str): string specifying the pennylane device type; e.g. 'default.qubit'
            qnode_kwargs (str): the key word arguments passed to the circuit qnode
        """

        # attributes that do not depend on data
        self.learning_rate = learning_rate
        self.max_steps = max_steps
        self.convergence_interval = convergence_interval
        self.batch_size = batch_size
        self.dev_type = dev_type
        self.jit = jit
        self.qnode_kwargs = qnode_kwargs
        self.scaling = scaling
        self.random_state = random_state
        self.rng = np.random.default_rng(random_state)

        if max_vmap is None:
            self.max_vmap = self.batch_size
        else:
            self.max_vmap = max_vmap

        # data-dependant attributes
        # which will be initialised by calling "fit"
        self.params_ = None  # Dictionary containing the trainable parameters
        self.circuit = None

    def generate_key(self):
        return jax.random.PRNGKey(self.rng.integers(1000000))

    def construct_model(self):
        dev = qml.device(self.dev_type, wires=self.n_qubits)
        n_layers = int(jnp.log2(self.n_qubits))

        @qml.qnode(dev, **self.qnode_kwargs)
        def circuit(params, x):
            """
            The tree tensor QNN from the plots
            """
            qml.AmplitudeEmbedding(x, normalize=True, wires=range(self.n_qubits))
            count = 0
            for layer in range(n_layers):
                for q in range(0, self.n_qubits, 2**layer):
                    qml.RY(params["weights"][count], wires=q)
                    count = count + 1
                qml.broadcast(
                    qml.CNOT,
                    wires=range(self.n_qubits),
                    pattern=[
                        ((i + 2**layer), i)
                        for i in range(0, self.n_qubits, 2 ** (layer + 1))
                    ],
                )
            qml.RY(params["weights"][count], wires=0)
            return qml.expval(qml.PauliZ(wires=0))

        self.circuit = circuit

        if self.jit:
            circuit = jax.jit(circuit)

        def circuit_plus_bias(params, x):
            return circuit(params, x) + params["bias"]

        self.forward = jax.vmap(circuit_plus_bias, in_axes=(None, 0))
        self.chunked_forward = chunk_vmapped_fn(self.forward, 1, self.max_vmap)

        return self.forward

    def initialize(self, n_features, classes=None):
        """Initialize attributes that depend on the number of features and the class labels.

        Args:
            n_features (int): Number of features that the classifier expects
            classes (array-like): class labels that the classifier expects
        """
        if classes is None:
            classes = [-1, 1]

        self.classes_ = classes
        self.n_classes_ = len(self.classes_)
        assert self.n_classes_ == 2
        assert 1 in self.classes_ and -1 in self.classes_

        if n_features == 1:
            self.n_qubits = 1
        else:
            n_qubits_ae = int(
                np.ceil(np.log2(n_features))
            )  # the num qubits needed to amplitude encode
            n_qubits = 2 ** int(
                np.ceil(np.log2(n_qubits_ae))
            )  # the model needs 2**m qubits, for some m
            self.n_qubits = n_qubits

        self.initialize_params()
        self.construct_model()

    def initialize_params(self):
        # initialise the trainable parameters
        weights = (
            2
            * jnp.pi
            * jax.random.uniform(
                shape=(2 * self.n_qubits - 1,), key=self.generate_key()
            )
        )

        bias = 0.1 * jax.random.normal(shape=(1,), key=self.generate_key())
        self.params_ = {"weights": weights, "bias": bias}

    def fit(self, X, y):
        """Fit the model to data X and labels y.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
            y (np.ndarray): Labels of shape (n_samples,)
        """
        self.initialize(X.shape[1], classes=np.unique(y))

        X = self.transform(X)

        def loss_fn(params, X, y):
            # square loss
            predictions = self.forward(params, X)
            return jnp.mean((predictions - y) ** 2)

        if self.jit:
            loss_fn = jax.jit(loss_fn)
        optimizer = optax.adam
        self.params_ = train(
            self,
            loss_fn,
            optimizer,
            X,
            y,
            self.generate_key,
            convergence_interval=self.convergence_interval,
        )

        return self

    def predict(self, X):
        """Predict labels for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred (np.ndarray): Predicted labels of shape (n_samples,)
        """
        predictions = self.predict_proba(X)
        mapped_predictions = np.argmax(predictions, axis=1)
        return np.take(self.classes_, mapped_predictions)

    def predict_proba(self, X):
        """Predict label probabilities for data X.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)

        Returns:
            y_pred_proba (np.ndarray): Predicted label probabilities of shape
            (n_samples, n_classes)
        """
        X = self.transform(X)
        predictions = self.chunked_forward(self.params_, X)
        predictions_2d = np.c_[(1 - predictions) / 2, (1 + predictions) / 2]
        return predictions_2d

    def transform(self, X, preprocess=True):
        """
        The feature vectors padded to the next power of 2 and then normalised.

        Args:
            X (np.ndarray): Data of shape (n_samples, n_features)
        """
        n_features = X.shape[1]
        X = X * self.scaling

        n_qubits_ae = int(
            np.ceil(np.log2(n_features))
        )  # the num qubits needed to amplitude encode
        n_qubits = 2 ** int(
            np.ceil(np.log2(n_qubits_ae))
        )  # the model needs 2**m qubits, for some m
        max_n_features = 2**n_qubits
        n_padding = max_n_features - n_features
        padding = np.ones(shape=(len(X), n_padding)) / max_n_features

        X_padded = np.c_[X, padding]
        X_normalised = np.divide(
            X_padded, np.expand_dims(np.linalg.norm(X_padded, axis=1), axis=1)
        )
        return X_normalised



# Copyright 2024 Xanadu Quantum Technologies Inc.
# Modified by Albert Nieto, 2024.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def generate_bars_and_stripes(n_samples, height, width, noise_std):
    """Data generation procedure for 'bars and stripes'.

    Args:
        n_samples (int): number of data samples to produce
        height (int): number of pixels for image height
        width (int): number of pixels for image width
        noise_std (float): standard deviation of Gaussian noise added to the pixels
    """
    X = np.ones([n_samples, 1, height, width]) * -1
    y = np.zeros([n_samples])

    for i in range(len(X)):
        if np.random.rand() > 0.5:
            rows = np.where(np.random.rand(width) > 0.5)[0]
            X[i, 0, rows, :] = 1.0
            y[i] = -1
        else:
            columns = np.where(np.random.rand(height) > 0.5)[0]
            X[i, 0, :, columns] = 1.0
            y[i] = +1
        X[i, 0] = X[i, 0] + np.random.normal(0, noise_std, size=X[i, 0].shape)

    return X, y



# Copyright 2024 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Score a model using the best hyperparameters, using a command-line script."""

import numpy as np
import sys
import os
import argparse
import logging
import pandas as pd

logging.getLogger().setLevel(logging.INFO)
from importlib import import_module
from pathlib import Path
from qcml.utils.hyperparam_search import read_data, csv_to_dict

np.random.seed(42)

logging.info("cpu count:" + str(os.cpu_count()))


if __name__ == "__main__":
    # Create an argument parser
    parser = argparse.ArgumentParser(
        description="Run experiments with hyperparameter search.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--classifier-name",
        help="Classifier to run",
    )

    parser.add_argument(
        "--trainset-path",
        help="Path to the training set",
    )

    parser.add_argument(
        "--testset-path",
        help="Path to the test set",
    )

    parser.add_argument(
        "--hyperparams-path",
        default=".",
        help="Path to the file with the best hyperparameters",
    )

    parser.add_argument(
        "--results-path", default=".", help="Path to store the experiment results"
    )

    parser.add_argument(
        "--clean",
        help="True or False. Remove previous results if it exists",
        dest="clean",
        default=False,
        type=bool,
    )

    parser.add_argument(
        "--n-jobs", type=int, default=-1, help="Number of parallel threads to run"
    )

    # Parse the arguments along with any extra arguments that might be model specific
    args, unknown_args = parser.parse_known_args()

    if any(
        arg is None
        for arg in [args.classifier_name, args.trainset_path, args.testset_path]
    ):
        msg = "\n================================================================================"
        msg += "\nA classifier from qml.benchmarks.model and dataset path are required. E.g., \n \n"
        msg += (
            "python score_with_best_hyperparameters \ \n"
            "--classifier DataReuploadingClassifier \ \n"
            "--trainset-path my_train_data.csv\n"
            "--testset-path my_test_data.csv\n"
        )
        msg += "\nCheck all arguments for the script with \n"
        msg += "python score_with_best_hyperparameters --help\n"
        msg += "================================================================================"
        raise ValueError(msg)

    experiment_path = args.results_path
    results_path = os.path.join(experiment_path, "results")

    if not os.path.exists(results_path):
        os.makedirs(results_path)

    ###################################################################
    # Get the classifier, dataset and best hyperparameters
    ###################################################################
    Classifier = getattr(import_module("qml_benchmarks.models"), args.classifier_name)
    classifier_name = Classifier.__name__

    # Load the data
    train_dataset_filename = os.path.join(args.trainset_path)
    X_train, y_train = read_data(train_dataset_filename)

    test_dataset_filename = os.path.join(args.testset_path)
    X_test, y_test = read_data(test_dataset_filename)

    # Construct output path
    dataset_path_obj = Path(args.trainset_path)
    results_filename_stem = " ".join(
        [Classifier.__name__ + "_" + dataset_path_obj.stem + "_GridSearchCV"]
    )

    # If we have already run this experiment then continue
    path_out = os.path.join(
        results_path, results_filename_stem + "-best-hyperparams-results.csv"
    )
    if os.path.isfile(path_out):
        if args.clean is False:
            msg = "\n================================================================================="
            msg += "\nResults exist in " + path_out
            msg += "\nSpecify --clean True to override results or new --results-path"
            msg += "\n================================================================================="
            logging.warning(msg)
            sys.exit(msg)
        else:
            logging.warning("Cleaning existing results for ", path_out)

    # Load best hyperparameters
    best_hyperparams = csv_to_dict(args.hyperparams_path)

    # Score the model
    results_with_best_hyperparams = {"train_acc": [], "test_acc": []}
    for i in range(5):
        classifier = Classifier(**best_hyperparams, random_state=i)
        classifier.fit(X_train, y_train)

        acc_train = classifier.score(X_train, y_train)
        acc_test = classifier.score(X_test, y_test)
        results_with_best_hyperparams["train_acc"].append(acc_train)
        results_with_best_hyperparams["test_acc"].append(acc_test)

    print("Results with best hyperparams", results_with_best_hyperparams)
    df = pd.DataFrame.from_dict(results_with_best_hyperparams)
    df.to_csv(os.path.join(path_out))



import numpy as np


def zero_state_array(n):
    """
    Create an n-qubit zero state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: Zero state for n qubits.
    """
    return np.kron(*[np.array([1, 0]) for _ in range(n)])


def one_state_array(n):
    """
    Create an n-qubit one state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: One state for n qubits.
    """
    return np.kron(*[np.array([0, 1]) for _ in range(n)])


def plus_state_array(n):
    """
    Create an n-qubit plus state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: Plus state for n qubits.
    """
    return np.kron(*[np.array([1, 1]) / np.sqrt(2) for _ in range(n)])


def minus_state_array(n):
    """
    Create an n-qubit minus state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: Minus state for n qubits.
    """
    return np.kron(*[np.array([1, -1]) / np.sqrt(2) for _ in range(n)])


def i_plus_state_array(n):
    """
    Create an n-qubit |i+⟩ state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: |i+⟩ state for n qubits.
    """
    return np.kron(
        *[(np.array([1, 0]) + 1j * np.array([0, 1])) / np.sqrt(2) for _ in range(n)]
    )


def i_minus_state_array(n):
    """
    Create an n-qubit |i-⟩ state.

    Parameters:
    - n (int): Number of qubits.

    Returns:
    - np.ndarray: |i-⟩ state for n qubits.
    """
    return np.kron(
        *[(np.array([1, 0]) - 1j * np.array([0, 1])) / np.sqrt(2) for _ in range(n)]
    )


def bell_state_1_array():
    """
    Create the first Bell state |Φ⁺⟩.

    Returns:
    - np.ndarray: |Φ⁺⟩ Bell state.
    """
    return (
        np.kron(np.array([1, 0]), np.array([1, 0]))
        + np.kron(np.array([0, 1]), np.array([0, 1]))
    ) / np.sqrt(2)


def bell_state_2_array():
    """
    Create the second Bell state |Φ⁻⟩.

    Returns:
    - np.ndarray: |Φ⁻⟩ Bell state.
    """
    return (
        np.kron(np.array([1, 0]), np.array([1, 0]))
        - np.kron(np.array([0, 1]), np.array([0, 1]))
    ) / np.sqrt(2)


def bell_state_3_array():
    """
    Create the third Bell state |Ψ⁺⟩.

    Returns:
    - np.ndarray: |Ψ⁺⟩ Bell state.
    """
    return (
        np.kron(np.array([1, 0]), np.array([0, 1]))
        + np.kron(np.array([0, 1]), np.array([1, 0]))
    ) / np.sqrt(2)


def bell_state_4_array():
    """
    Create the fourth Bell state |Ψ⁻⟩.

    Returns:
    - np.ndarray: |Ψ⁻⟩ Bell state.
    """
    return (
        np.kron(np.array([1, 0]), np.array([0, 1]))
        - np.kron(np.array([0, 1]), np.array([1, 0]))
    ) / np.sqrt(2)



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from IPython.core.magic import register_line_magic
from IPython.display import display, HTML


@register_line_magic
def scroll_output(line):
    """Magic command to enable scrolling in output cells with a specified max height."""
    height = line.strip() or "300px"
    display(
        HTML(
            f"""
        <style>
            .output_area pre {{
                max-height: {height};
                overflow-y: scroll;
            }}
            .output_area {{
                max-height: {height};
                overflow-y: scroll;
            }}
        </style>
    """
        )
    )


# Now you can use the magic command in your notebook



# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from qcml.kernels.jax import (
    sigmoid_kernel,
    laplacian_kernel,
    anova_kernel,
    chi_squared_kernel,
    histogram_intersection_kernel,
    linear_kernel,
    polynomial_kernel,
    gaussian_kernel,
    rbf_kernel,
)
from qcml.kernels import (
    separable_kernel,
    projected_quantum_kernel,
    iqp_embedding_kernel,
    angle_embedding_kernel,
    amplitude_embedding_kernel,
)

classical_kernel_param_map = {
    "sigmoid_kernel": {"alpha", "c"},
    "laplacian_kernel": {"gamma"},
    "anova_kernel": {"d", "sigma"},
    "chi_squared_kernel": set(),
    "histogram_intersection_kernel": set(),
    "linear_kernel": set(),
    "polynomial_kernel": {"degree", "coef0"},
    "gaussian_kernel": {"gamma"},
    "rbf_kernel": {"gamma"},
}


# Define the parameter grids for each kernel function
classical_kernel_grid = (
    [
        {"kernel_func": sigmoid_kernel, "kernel_params": {"alpha": alpha, "c": c}}
        for alpha, c in itertools.product([0.1, 0.5, 1.0], [0, 1, 5])
    ]
    + [
        {"kernel_func": laplacian_kernel, "kernel_params": {"gamma": gamma}}
        for gamma in [0.1, 1.0, 10.0]
    ]
    + [
        {"kernel_func": anova_kernel, "kernel_params": {"d": d, "sigma": sigma}}
        for d, sigma in itertools.product([2, 3], [0.1, 1.0])
    ]
    + [{"kernel_func": chi_squared_kernel, "kernel_params": {}}]
    + [{"kernel_func": histogram_intersection_kernel, "kernel_params": {}}]
    + [{"kernel_func": linear_kernel, "kernel_params": {}}]
    + [
        {
            "kernel_func": polynomial_kernel,
            "kernel_params": {"degree": degree, "coef0": coef0},
        }
        for degree, coef0 in itertools.product([2, 3, 4], [0, 1])
    ]
    + [
        {"kernel_func": gaussian_kernel, "kernel_params": {"gamma": gamma}}
        for gamma in [0.1, 1.0, 10.0]
    ]
    + [
        {"kernel_func": rbf_kernel, "kernel_params": {"gamma": gamma}}
        for gamma in [0.1, 1.0, 10.0]
    ]
)

reduced_classical_kernel_grid = (
    [
        {"kernel_func": sigmoid_kernel, "kernel_params": {"alpha": alpha, "c": c}}
        for alpha, c in itertools.product([0.1, 0.5, 1.0], [0, 1, 5])
    ]
    + [
        {
            "kernel_func": polynomial_kernel,
            "kernel_params": {"degree": degree, "coef0": coef0},
        }
        for degree, coef0 in itertools.product([2, 3, 4], [0, 1])
    ]
    + [
        {"kernel_func": rbf_kernel, "kernel_params": {"gamma": gamma}}
        for gamma in [0.1, 1.0, 10.0]
    ]
)

quantum_kernel_param_map = {
    "separable_kernel": {"encoding_layers"},
    "projected_quantum_kernel": {"trotter_steps", "t", "gamma_factor", "embedding"},
    "iqp_embedding_kernel": {"repeats"},
    "angle_embedding_kernel": {"rotation"},
    "amplitude_embedding_kernel": set(),
}


quantum_kernel_grid = (
    [
        {
            "kernel_func": projected_quantum_kernel,
            "kernel_params": {
                "trotter_steps": trotter_steps,
                "t": t,
                "gamma_factor": gamma_factor,
                "embedding": embedding,
            },
        }
        for trotter_steps, t, gamma_factor, embedding in itertools.product(
            [1, 3, 5], [0.01, 0.1, 1], [0.1, 1, 10], ["Hamiltonian", "IQP"]
        )
    ]
    + [
        {
            "kernel_func": separable_kernel,
            "kernel_params": {"encoding_layers": encoding_layers},
        }
        for encoding_layers in [1, 2, 3]
    ]
    + [
        {"kernel_func": iqp_embedding_kernel, "kernel_params": {"repeats": repeats}}
        for repeats in [1, 2, 3]
    ]
    + [
        {"kernel_func": angle_embedding_kernel, "kernel_params": {"rotation": rotation}}
        for rotation in ["X", "Y", "Z"]
    ]
    + [{"kernel_func": amplitude_embedding_kernel, "kernel_params": {}}]
)

kernel_grid = classical_kernel_grid + quantum_kernel_grid
kernel_param_map = {**classical_kernel_param_map, **quantum_kernel_param_map}

reduced_quantum_kernel_grid = [
    {
        "kernel_func": projected_quantum_kernel,
        "kernel_params": {
            "trotter_steps": trotter_steps,
            "t": t,
            "gamma_factor": gamma_factor,
        },
    }
    for trotter_steps, t, gamma_factor in itertools.product(
        [1, 3, 5], [0.01, 0.1, 1], [0.1, 1, 10]
    )
] + [
    {"kernel_func": iqp_embedding_kernel, "kernel_params": {"repeats": repeats}}
    for repeats in [1, 2, 3]
]
reduced_kernel_grid = (
    [
        {
            "kernel_func": projected_quantum_kernel,
            "kernel_params": {
                "trotter_steps": trotter_steps,
                "t": t,
                "gamma_factor": gamma_factor,
            },
        }
        for trotter_steps, t, gamma_factor in itertools.product(
            [1, 3, 5], [0.01, 0.1, 1], [0.1, 1, 10]
        )
    ]
    + [
        {
            "kernel_func": separable_kernel,
            "kernel_params": {"encoding_layers": encoding_layers},
        }
        for encoding_layers in [1, 2, 3]
    ]
    + [
        {"kernel_func": rbf_kernel, "kernel_params": {"gamma": gamma}}
        for gamma in [0.1, 1.0, 10.0]
    ]
)
more_reduced_kernel_grid = reduced_classical_kernel_grid + reduced_quantum_kernel_grid


def select_kernels(selected_kernels, kernel_grid=kernel_grid):
    """
    Filter the kernel grid to include only selected kernels.

    Parameters:
    - kernel_grid (list): The complete list of kernel configurations.
    - selected_kernels (list): The list of kernel function names to include.

    Returns:
    - filtered_grid (list): The filtered list of kernel configurations.
    """
    filtered_grid = [
        kernel
        for kernel in kernel_grid
        if kernel["kernel_func"].__name__ in selected_kernels
    ]
    return filtered_grid
